@@ -4,6 +4,7 @@
 
 import io
 import logging
+import math
 
 from smbclient._pool import (
     ClientConfig,
@@ -34,7 +35,6 @@ from smbprotocol.ioctl import (
 from smbprotocol.file_info import (
     FileAttributes,
     FileEndOfFileInformation,
-    FileStandardInformation,
 )
 
 from smbprotocol.open import (
@@ -45,6 +45,7 @@ from smbprotocol.open import (
     Open,
     QueryDirectoryFlags,
     ShareAccess,
+    SMB2CreateRequest,
     SMB2QueryInfoRequest,
     SMB2QueryInfoResponse,
     SMB2SetInfoRequest,
@@ -110,6 +111,66 @@ def _parse_mode(raw, invalid=""):
     return create_disposition
 
 
+def _chunk_size(connection, length, operation):
+    """
+    Get the maximum size of data we can read/write. Also gets the number of credits to request to optimize subsequent
+    read/write operations for the remaining length.
+
+    :param connection: The SMB connection.
+    :param length: The length of the data we are working with.
+    :param operation: The operation the chunk is for: 'read', 'write', 'transact'
+    :return: The size of the chunk we can use and the number of credits to request for the next operation.
+    """
+    max_size = getattr(connection, 'max_%s_size' % operation)
+
+    # Determine the maximum data length we can send for the operation. We do this by checking the available credits and
+    # calculating whatever is the smallest; length, negotiated operation size, available credit size).
+    available_credits = connection.sequence_window['high'] - connection.sequence_window['low']
+    chunk_size = min(length, max_size, available_credits * MAX_PAYLOAD_SIZE)
+
+    # Determine how many credits we need to fully optimize subsequent calls for the remaining amount of data. Basically
+    # how many credits we need to send either the remaining data or the max operation size.
+    remaining_length = min(max(0, length - chunk_size), max_size)
+    consumed_credits = math.ceil(chunk_size / MAX_PAYLOAD_SIZE)
+    credit_request = consumed_credits + math.ceil(remaining_length / MAX_PAYLOAD_SIZE)
+
+    return chunk_size, int(credit_request)
+
+
+def _resolve_dfs(raw_io):
+    """
+    Resolves a DFS path for a failed Open request.
+
+    :param raw_io: The SMBRawIO to resolve the DFS path for.
+    :return: A new Open for each DFS target that was resolved.
+    """
+    if not raw_io.fd.tree_connect.is_dfs_share:
+        return
+
+    # Path is on a DFS root that is linked to another server.
+    client_config = ClientConfig()
+    raw_path = raw_io.name
+
+    referral = dfs_request(raw_io.fd.tree_connect, raw_path[1:])
+    client_config.cache_referral(referral)
+    info = client_config.lookup_referral([p for p in raw_path.split("\\") if p])
+    connection_kwargs = getattr(raw_io, '_%s__kwargs' % type(raw_io).__name__, {})
+
+    for target in info:
+        new_path = raw_path.replace(info.dfs_path, target.target_path, 1)
+
+        try:
+            tree, fd_path = get_smb_tree(new_path, **connection_kwargs)
+
+        except SMBResponseException as link_exc:
+            log.warning("Failed to connect to DFS link target %s: %s" % (str(target), link_exc))
+            continue
+
+        # Record the target that worked for future reference.
+        info.target_hint = target
+        yield Open(tree, fd_path)
+
+
 def ioctl_request(transaction, ctl_code, output_size=0, flags=IOCTLFlags.SMB2_0_IOCTL_IS_IOCTL, input_buffer=b""):
     """
     Sends an IOCTL request to the server.
@@ -214,17 +275,21 @@ class SMBFileTransaction(object):
         """
         Sends a compound request to the server. Optionally opens and closes a handle in the same request if the handle
         is not already opened.
-
-        :return: A list of each response for the requests set on the transaction.
         """
         remove_index = []
-        if self.raw.closed:
+        if (
+                self.raw.closed and
+                (
+                        not self._actions or
+                        not isinstance(self._actions[0][0], SMB2CreateRequest)
+                )
+        ):
             self.raw.open(transaction=self)
             self._actions.insert(0, self._actions.pop(-1))  # Need to move the create to the start
-            remove_index.append(0)
+            remove_index.insert(0, 0)
 
             self.raw.close(transaction=self)
-            remove_index.append(len(self._actions) - 1)
+            remove_index.insert(0, len(self._actions) - 1)
 
         send_msgs = []
         unpack_functions = []
@@ -240,19 +305,61 @@ class SMBFileTransaction(object):
         # just enumerate the list in 1 line in case it throws an exception.
         failures = []
         responses = []
+        try_again = False
         for idx, func in enumerate(unpack_functions):
+            res = None
+
             try:
-                responses.append(func(requests[idx]))
+                try:
+                    res = func(requests[idx])
+
+                except (PathNotCovered, ObjectNameNotFound, ObjectPathNotFound):
+                    # The MS-DFSC docs state that STATUS_PATH_NOT_COVERED is used when encountering a link to a
+                    # different server but Samba seems to return the generic name or path not found.
+
+                    # If the first action is not a CreateRequest then we can't resolve the DFS path in the transaction.
+                    if not (idx == 0 and isinstance(send_msgs[0], SMB2CreateRequest)):
+                        raise
+
+                    for smb_open in _resolve_dfs(self.raw):
+                        if smb_open.tree_connect.share_name == self.raw.fd.tree_connect.share_name:
+                            continue
+
+                        self.raw.fd = smb_open
+
+                        # In case this is a transaction with an explicit open we want to reopen it with the new params
+                        # before trying it again.
+                        self.raw.open(transaction=self)
+                        self._actions[0] = self._actions.pop(-1)
+
+                        try_again = True
+                        break
+
+                    else:
+                        # Either there wasn't any DFS referrals or none of them worked, just reraise the error.
+                        raise
+
             except SMBResponseException as exc:
                 failures.append(SMBOSError(exc.status, self.raw.name))
 
-        # If there was a failure, just raise the first exception found.
-        if failures:
+            finally:
+                responses.append(res)
+
+        # Remove the existing open/close actions this transaction added internally.
+        for idx in remove_index:
+            del self._actions[idx]
+            del responses[idx]
+
+        if try_again:
+            # If we updated the SMB Tree due to a DFS referral hit then try again.
+            self.commit()
+
+        elif failures:
+            # If there was a failure, just raise the first exception found.
             raise failures[0]
 
-        remove_index.sort(reverse=True)  # Need to remove from highest index first.
-        [responses.pop(i) for i in remove_index]  # Remove the open and close response if commit() did that work.
-        self.results = tuple(responses)
+        else:
+            self.results = tuple(responses)
 
 
 class SMBRawIO(io.RawIOBase):
@@ -261,7 +368,7 @@ class SMBRawIO(io.RawIOBase):
     _INVALID_MODE = ''
 
     def __init__(self, path, mode='r', share_access=None, desired_access=None, file_attributes=None,
-                 create_options=0, buffer_size=MAX_PAYLOAD_SIZE, **kwargs):
+                 create_options=0, **kwargs):
         tree, fd_path = get_smb_tree(path, **kwargs)
         self.share_access = share_access
         self.fd = Open(tree, fd_path)
@@ -269,7 +376,6 @@ class SMBRawIO(io.RawIOBase):
         self._name = path
         self._offset = 0
         self._flush = False
-        self._buffer_size = buffer_size
         self.__kwargs = kwargs  # Used in open for DFS referrals
 
         if desired_access is None:
@@ -333,63 +439,32 @@ class SMBRawIO(io.RawIOBase):
         if not self.closed:
             return
 
+        open_at_end = False
+        if not transaction:
+            # The SMBFileTransaction has special logic that deals with DFS paths, use that to do the actual open.
+            open_at_end = True
+            transaction = SMBFileTransaction(self)
+
         share_access = _parse_share_access(self.share_access, self.mode)
         create_disposition = _parse_mode(self.mode, invalid=self._INVALID_MODE)
 
-        try:
-            # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wpo/feeb3122-cfe0-4b34-821d-e31c036d763c
-            # Impersonation on SMB has little meaning when opening files but is important if using RPC so set to a sane
-            # default of Impersonation.
-            open_result = self.fd.create(
-                ImpersonationLevel.Impersonation,
-                self._desired_access,
-                self._file_attributes,
-                share_access,
-                create_disposition,
-                self._create_options,
-                send=(transaction is None),
-            )
-        except (PathNotCovered, ObjectNameNotFound, ObjectPathNotFound) as exc:
-            # The MS-DFSC docs status that STATUS_PATH_NOT_COVERED is used when encountering a link to a different
-            # server but Samba seems to return the generic name or path not found.
-            if not self.fd.tree_connect.is_dfs_share:
-                raise SMBOSError(exc.status, self.name)
+        # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wpo/feeb3122-cfe0-4b34-821d-e31c036d763c
+        # Impersonation on SMB has little meaning when opening files but is important if using RPC so set to a sane
+        # default of Impersonation.
+        transaction += self.fd.create(
+            ImpersonationLevel.Impersonation,
+            self._desired_access,
+            self._file_attributes,
+            share_access,
+            create_disposition,
+            self._create_options,
+            send=False,
+        )
 
-            # Path is on a DFS root that is linked to another server.
-            client_config = ClientConfig()
-            referral = dfs_request(self.fd.tree_connect, self.name[1:])
-            client_config.cache_referral(referral)
-            info = client_config.lookup_referral([p for p in self.name.split("\\") if p])
-
-            for target in info:
-                new_path = self.name.replace(info.dfs_path, target.target_path, 1)
-
-                try:
-                    tree, fd_path = get_smb_tree(new_path, **self.__kwargs)
-                    self.fd = Open(tree, fd_path)
-                    self.open(transaction=transaction)
-
-                except SMBResponseException as link_exc:
-                    log.warning("Failed to connect to DFS link target %s: %s" % (str(target), link_exc))
-
-                else:
-                    # Record the target that worked for future reference.
-                    info.target_hint = target
-                    break
-
-            else:
-                # None of the targets worked so raise the original error.
-                raise SMBOSError(exc.status, self.name)
-
-            return
-
-        except SMBResponseException as exc:
-            raise SMBOSError(exc.status, self.name)
-
-        if transaction is not None:
-            transaction += open_result
-        elif 'a' in self.mode and self.FILE_TYPE != 'pipe':
-            self._offset = self.fd.end_of_file
+        if open_at_end:
+            transaction.commit()
+            if 'a' in self.mode and self.FILE_TYPE != 'pipe':
+                self._offset = self.fd.end_of_file
 
     def readable(self):
         """ True if file was opened in a read mode. """
@@ -454,24 +529,29 @@ class SMBRawIO(io.RawIOBase):
 
         :return: The byte string read from the SMB file.
         """
-        data = b""
-        remaining_bytes = self.fd.end_of_file - self._offset
-        while len(data) < remaining_bytes or self.FILE_TYPE == 'pipe':
-            try:
-                data_part = self.fd.read(self._offset, self._buffer_size)
-            except PipeBroken:
+        data = bytearray()
+        while True:
+            read_length = min(
+                # We always want to be reading a minimum of 64KiB.
+                max(self.fd.end_of_file - self._offset, MAX_PAYLOAD_SIZE),
+                self.fd.connection.max_read_size  # We can never read more than this.
+            )
+
+            buffer = bytearray(b'\x00' * read_length)
+            bytes_read = self.readinto(buffer)
+            if not bytes_read:
                 break
 
-            data += data_part
-            if self.FILE_TYPE != 'pipe':
-                self._offset += len(data_part)
+            data += buffer[:bytes_read]
 
-        return data
+        return bytes(data)
 
     def readinto(self, b):
         """
         Read bytes into a pre-allocated, writable bytes-like object b, and
-        return the number of bytes read.
+        return the number of bytes read. This may read less bytes than
+        requested as it depends on the negotiated read size and SMB credits
+        available.
 
         :param b: bytes-like object to read the data into.
         :return: The number of bytes read.
@@ -479,9 +559,20 @@ class SMBRawIO(io.RawIOBase):
         if self._offset >= self.fd.end_of_file and self.FILE_TYPE != 'pipe':
             return 0
 
+        chunk_size, credit_request = _chunk_size(self.fd.connection, len(b), 'read')
+
+        read_msg, recv_func = self.fd.read(self._offset, chunk_size, send=False)
+        request = self.fd.connection.send(
+            read_msg,
+            sid=self.fd.tree_connect.session.session_id,
+            tid=self.fd.tree_connect.tree_connect_id,
+            credit_request=credit_request
+        )
+
         try:
-            file_bytes = self.fd.read(self._offset, len(b))
+            file_bytes = recv_func(request)
         except PipeBroken:
+            # A pipe will block until it returns the data available or was closed/broken.
             file_bytes = b""
 
         b[:len(file_bytes)] = file_bytes
@@ -496,23 +587,29 @@ class SMBRawIO(io.RawIOBase):
         Write buffer b to file, return number of bytes written.
 
         Only makes one system call, so not all of the data may be written.
-        The number of bytes actually written is returned.
+        The number of bytes actually written is returned. This can be less than
+        the length of b as it depends on the underlying connection.
         """
+        chunk_size, credit_request = _chunk_size(self.fd.connection, len(b), 'write')
+
+        # Python 2 compat, can be removed and just use the else statement.
         if isinstance(b, memoryview):
-            b = b.tobytes()
+            data = b[:chunk_size].tobytes()
+        else:
+            data = bytes(b[:chunk_size])
 
-        with SMBFileTransaction(self) as transaction:
-            transaction += self.fd.write(b, offset=self._offset, send=False)
+        write_msg, recv_func = self.fd.write(data, offset=self._offset, send=False)
+        request = self.fd.connection.send(
+            write_msg,
+            sid=self.fd.tree_connect.session.session_id,
+            tid=self.fd.tree_connect.tree_connect_id,
+            credit_request=credit_request
+        )
+        bytes_written = recv_func(request)
 
-            # Send the request with an SMB2QueryInfoRequest for FileStandardInformation so we can update the end of
-            # file stored internally.
-            if self.FILE_TYPE != 'pipe':
-                query_info(transaction, FileStandardInformation)
-
-        bytes_written = transaction.results[0]
         if self.FILE_TYPE != 'pipe':
             self._offset += bytes_written
-            self.fd.end_of_file = transaction.results[1]['end_of_file'].get_value()
+            self.fd.end_of_file = max(self.fd.end_of_file, self._offset)
             self._flush = True
 
         return bytes_written
