@@ -52,6 +52,11 @@ class SMBAuthenticationError(SMBException):
     pass
 
 
+class SMBConnectionClosed(SMBException):
+    # Used to denote the underlying TCP transport has been closed.
+    pass
+
+
 class SMBOSError(OSError, SMBException):
     """Wrapper for OSError with smbprotocol specific details.
 
@@ -72,6 +77,7 @@ class SMBOSError(OSError, SMBException):
         error_details = {
             NtStatus.STATUS_OBJECT_NAME_NOT_FOUND: errno.ENOENT,
             NtStatus.STATUS_OBJECT_PATH_NOT_FOUND: errno.ENOENT,
+            NtStatus.STATUS_NOT_FOUND: errno.ENOENT,
             NtStatus.STATUS_OBJECT_NAME_COLLISION: errno.EEXIST,
             NtStatus.STATUS_PRIVILEGE_NOT_HELD: (errno.EACCES, "Required privilege not held"),
             NtStatus.STATUS_SHARING_VIOLATION: (errno.EPERM, "The process cannot access the file because it is being "
@@ -347,6 +353,11 @@ class NoSuchFile(SMBResponseException):
     _STATUS_CODE = NtStatus.STATUS_NO_SUCH_FILE
 
 
+class InvalidDeviceRequest(SMBResponseException):
+    _BASE_MESSAGE = "The specified request is not a valid operation for the target device."
+    _STATUS_CODE = NtStatus.STATUS_INVALID_DEVICE_REQUEST
+
+
 class MoreProcessingRequired(SMBResponseException):
     _BASE_MESSAGE = "The specified I/O request packet (IRP) cannot be disposed of because the I/O operation is not " \
                     "complete."
@@ -428,6 +439,11 @@ class PrivilegeNotHeld(SMBResponseException):
     _STATUS_CODE = NtStatus.STATUS_PRIVILEGE_NOT_HELD
 
 
+class WrongPassword(SMBResponseException):
+    _BASE_MESSAGE = "The specified password is not correct or the user is locked out."
+    _STATUS_CODE = NtStatus.STATUS_WRONG_PASSWORD
+
+
 class LogonFailure(SMBResponseException):
     _BASE_MESSAGE = "The attempted logon is invalid. This is either due to a bad username or authentication " \
                     "information."
